'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.AggregateIndex = undefined;

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _rheactorErrors = require('@resourcefulhumans/rheactor-errors');

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var AggregateIndex = exports.AggregateIndex = function () {
  /**
   * Manages indices for aggregates
   *
   * @param {String} aggregate
   * @param {redis.client} redis
   */
  function AggregateIndex(aggregate, redis) {
    _classCallCheck(this, AggregateIndex);

    this.aggregate = aggregate;
    this.redis = redis;
  }

  /**
   * Add the value to the index of the given type for the aggregate identified by the given aggregateId
   *
   * @param {String} type
   * @param {String} value
   * @param {String} aggregateId
   * @returns {Promise}
   */


  _createClass(AggregateIndex, [{
    key: 'add',
    value: function add(type, value, aggregateId) {
      return this.redis.hmsetAsync(this.aggregate + '.' + type + '.index', value, aggregateId);
    }

    /**
     * Add the value to the index of the given type for the aggregate identified by the given aggregateId
     * if it is not already present
     *
     * @param {String} type
     * @param {String} value
     * @param {String} aggregateId
     * @returns {Promise}
     * @throws EntryAlreadyExistsError If entry exists
     */

  }, {
    key: 'addIfNotPresent',
    value: function addIfNotPresent(type, value, aggregateId) {
      var index = this.aggregate + '.' + type + '.index';
      return this.redis.evalAsync('local v = redis.call(\'HMGET\',ARGV[1],ARGV[2]) if v[1] == false then redis.call(\'HMSET\',ARGV[1],ARGV[2],ARGV[3]) return true else return false end', 0, index, value, aggregateId).then(function (res) {
        if (res === null) {
          throw new _rheactorErrors.EntryAlreadyExistsError('Entry for index "' + index + '" with key "' + value + '" already ' + 'exists. Tried to add aggregate "' + aggregateId + '".');
        }
        return true;
      });
    }

    /**
     * Remove the aggregate identified by the given aggregateId from the index of the given type for the given value
     *
     * @param {String} type
     * @param {String} value
     * @param {String} aggregateId
     * @returns {Promise}
     */

  }, {
    key: 'remove',
    value: function remove(type, value, aggregateId) {
      var index = this.aggregate + '.' + type + '.index';
      return this.redis.hdelAsync(index, value, aggregateId);
    }

    /**
     * Add the value to the list of the given type for the aggregate identified by the given aggregateId
     * if it is not already present
     *
     * @param {String} type
     * @param {String} aggregateId
     * @returns {Promise}
     * @throws EntryAlreadyExistsError If entry exists
     */

  }, {
    key: 'addToListIfNotPresent',
    value: function addToListIfNotPresent(type, aggregateId) {
      var index = this.aggregate + '.' + type + '.list';
      return this.redis.saddAsync(index, aggregateId).then(function (res) {
        if (!res) {
          throw new _rheactorErrors.EntryAlreadyExistsError('Aggregate "' + aggregateId + '" already member of "' + index + '".');
        }
        return true;
      });
    }

    /**
     * Returns the entries of the list
     *
     * @param {String} type
     * @returns {Promise.<Array>}
     */

  }, {
    key: 'getList',
    value: function getList(type) {
      var index = this.aggregate + '.' + type + '.list';
      return this.redis.smembersAsync(index);
    }

    /**
     * Remove the value from the list of the given type for the aggregate identified by the given aggregateId
     *
     * @param {String} type
     * @param {String} aggregateId
     * @returns {Promise}
     * @throws EntryAlreadyExistsError If entry exists
     */

  }, {
    key: 'removeFromList',
    value: function removeFromList(type, aggregateId) {
      var index = this.aggregate + '.' + type + '.list';
      return this.redis.sremAsync(index, aggregateId);
    }

    /**
     * Find an aggregateId by the given index type and value
     *
     * @param {String} type
     * @param {String} value
     * @returns {Promise}
     */

  }, {
    key: 'find',
    value: function find(type, value) {
      return this.get(type, value).catch(function (err) {
        return _rheactorErrors.EntryNotFoundError.is(err);
      }, function () {
        return null;
      });
    }

    /**
     * Return an aggregateId by the given index type and value
     *
     * @param {String} type
     * @param {String} value
     * @returns {Promise}
     * @throws EntryNotFoundError
     */

  }, {
    key: 'get',
    value: function get(type, value) {
      return this.redis.hmgetAsync(this.aggregate + '.' + type + '.index', value).then(function (res) {
        if (res[0] === null) {
          throw new _rheactorErrors.EntryNotFoundError('Aggregate not found with ' + type + ' "' + value + '"');
        }
        return res[0];
      });
    }

    /**
     * Return all aggregateIds in the given index type
     *
     * @param {String} type
     * @returns {Promise}
     */

  }, {
    key: 'getAll',
    value: function getAll(type) {
      return this.redis.hvalsAsync(this.aggregate + '.' + type + '.index');
    }
  }]);

  return AggregateIndex;
}();