'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.EventStore = undefined;

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _bluebird = require('bluebird');

var _modelEvent = require('./model-event');

var _types = require('./types');

var _tcomb = require('tcomb');

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

var EventStore = exports.EventStore = function () {
  /**
   * Creates a new EventStore for the given aggregate, e.g. 'user'.
   *
   * An event store maintains a version counter per aggregate id which guarantees that events have an ever increasing
   * version id.
   *
   * @param {String} aggregate
   * @param {redis.client} redis
   * @param {Number} numEvents
   */
  function EventStore(aggregate, redis) {
    var numEvents = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : 100;

    _classCallCheck(this, EventStore);

    (0, _tcomb.String)(aggregate);
    (0, _tcomb.Object)(redis);
    (0, _types.PositiveIntegerType)(numEvents);
    this.aggregate = aggregate;
    this.redis = redis;
    this.numEvents = numEvents;
  }

  /**
   * Persists an event
   *
   * The redis list type guarantees the order of insertion. So we don't need to jump through hoops to manage a version
   * id per aggregate. This can simply be done in the fetch method.
   *
   * @param {ModelEvent} event
   * @return {Promise}
   */


  _createClass(EventStore, [{
    key: 'persist',
    value: function persist(event) {
      (0, _modelEvent.ModelEventType)(event);
      var aggregateEvents = this.aggregate + '.events.' + event.aggregateId;
      var data = {
        eventType: event.name,
        eventPayload: event.data,
        eventCreatedAt: event.createdAt.getTime()
      };
      if (event.createdBy) {
        data.eventCreatedBy = event.createdBy;
      }
      return _bluebird.Promise.resolve(this.redis.rpushAsync(aggregateEvents, JSON.stringify(data)));
    }

    /**
     * Returns the events for the aggregate identified by aggregateId.
     *
     * @param {String} aggregateId
     * @return {Promise.<Array.<ModelEvent>>}
     */

  }, {
    key: 'fetch',
    value: function fetch(aggregateId) {
      var _this = this;

      (0, _types.AggregateIdType)(aggregateId);
      var aggregateEvents = this.aggregate + '.events.' + aggregateId;
      var start = 0;
      var fetchedEvents = [];
      var fetchEvents = function fetchEvents(start) {
        return _bluebird.Promise.resolve(_this.redis.lrangeAsync(aggregateEvents, start, start + _this.numEvents - 1)).then(function (res) {
          return _bluebird.Promise.map(res, function (e) {
            fetchedEvents.push(e);
          }).then(function () {
            if (res.length === _this.numEvents) {
              return fetchEvents(start + _this.numEvents);
            } else {
              return fetchedEvents;
            }
          });
        });
      };
      return fetchEvents(start).map(function (e) {
        var event = JSON.parse(e);
        var createdAt = new Date(event.eventCreatedAt || 0); // events did not always have this
        return new _modelEvent.ModelEvent(aggregateId, event.eventType, event.eventPayload, createdAt, event.eventCreatedBy);
      });
    }
  }]);

  return EventStore;
}();